#include <v_lib.h>
#include "eval.h"
#include "iplmath.h"
#include "ipltrans.h"
#include "decomp.h"
#include "cspace.h"
#include "messages.h"
#include "istrings.h"

/* *********************************************************** */

#define	MAT_SAME_SIZE(m1,m2) \
	((m1)->rows == (m2)->rows && (m1)->cols == (m2)->cols)

#define	VEC_SAME_SIZE(v1,v2) \
	((v1)->size == (v2)->size)

/* *********************************************************** */

vp_stack_t	cached_args = TYPED_STACK_EMPTY_INITIALIZER;

V_LOCAL arg_p
evalSearchForName(char *id)
{
	u32_t i, count = vpsCount(&cached_args);
	arg_p *args = (arg_p*)vpsGetStorage(&cached_args);
	for(i=0;i<count;i++)
		if(args[i] && STR_EQ(args[i]->id,id))
			return args[i];
	return NULL;
}

V_LOCAL void
evalFreeArg(arg_p arg)
{
	if(arg->id)
		memFree(arg->id);
	if(arg->owner && arg->data.any)
		arg->destroy(arg->data.any);
	memFree(arg);
}

void
evalFreeAllArgs(void)
{
	u32_t i, count = vpsCount(&cached_args);
	arg_p *args = (arg_p*)vpsGetStorage(&cached_args);
	for(i=0;i<count;i++)
		if(args[i])
			evalFreeArg(args[i]);
	vpsClear(&cached_args);
}

/* *********************************************************** */

vboolean
evalColorChannel(vhandle_t script,matrix_p channels[],
	u32_t index,
	int source_space,int assumed_space)
{
	CALLER_NAME(evalColorChannel)
	arg_p arg;
	char *name;
	if(source_space != assumed_space)
	{
		PARSE_ERROR(script,ipl_wrong_colorspace);
		return false;
	}
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
    name = strDup(lexLast(script));
    arg = evalArgument(name,ARG_MATRIX,script);
	memFree(name);
	if(!arg)
		return false;
	channels[index] = arg->data.matrix;
	return true;
}

vboolean
evalOutput(char *file_name,vhandle_t script)
{
	CALLER_NAME(evalOutput)
	char *name = NULL;
   	eval_state_t state;
	u32_t source_space = CSPACE_RGB;
	matrix_p channels[4] = {NULL,NULL,NULL,NULL};
	if(!lexNextIs(script,ipl_open_brace))
	{
		SYN_EXPECTED(script,ipl_open_brace);
		return false;
	}
	stateDefault(&state);
	while(lexNext(script))
    {
    	if(name)
    		memFree(name);
    	name = strDup(lexLast(script));
    	if(STR_EQ(name,ipl_close_brace))
        	break;
        else if(STR_EQ(name,ipl_key_state))
        {
			if(!stateChange(&state,script))
				goto err_exit;
        }
		else if(STR_EQ(name,ipl_key_source_space))
        {
        	char *space = lexNext(script);
            V_ASSERT(space);
        	if(STR_EQ(space,ipl_key_rgb))
            	source_space = CSPACE_RGB;
			else if(STR_EQ(space,ipl_key_hsv))
            	source_space = CSPACE_HSV;
			else if(STR_EQ(space,ipl_key_cie))
            	source_space = CSPACE_CIE;
			else
            {
            	PARSE_ERROR(script,ipl_wrong_colorspace);
                goto err_exit;
            }
        }
        else if(STR_EQ(name,ipl_key_red))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_RGB))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_green))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_RGB))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_blue))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_RGB))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_alpha))
        {
        	if(!evalColorChannel(script,channels,3,CSPACE_RGB,CSPACE_RGB))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_hue))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_HSV))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_saturation))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_HSV))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_value))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_HSV))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_cieX))
        {
        	if(!evalColorChannel(script,channels,0,source_space,CSPACE_CIE))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_cieY))
        {
        	if(!evalColorChannel(script,channels,1,source_space,CSPACE_CIE))
        		goto err_exit;
        }
        else if(STR_EQ(name,ipl_key_cieZ))
        {
        	if(!evalColorChannel(script,channels,2,source_space,CSPACE_CIE))
        		goto err_exit;
        }
        else
        {
        	PARSE_ERROR(script,ipl_unknown_channel)
       		goto err_exit;
        }
    }
    if(channels[0] && channels[1] && channels[2])
    {
    	if(channels[0]->rows != channels[1]->rows ||
    		channels[0]->rows != channels[2]->rows ||
    		channels[0]->cols != channels[1]->cols ||
    		channels[0]->cols != channels[2]->cols)
    	{
    		PARSE_ERROR(script,ipl_size_mismatch);
    		goto err_exit;
    	}
		outputSaveChannels(file_name,source_space,channels,&state);
		if(name)
			memFree(name);
	    return true;
    }
    PARSE_ERROR(script,ipl_not_enought_channels);
err_exit:;
	if(name)
		memFree(name);
	return false;
}

vboolean
evalAll(vhandle_t script)
{
	CALLER_NAME(evalAll)
	vboolean result = false;
	char *name = NULL;
	while(lexSkip_r(script,ipl_key_output))
	{
		if(!lexNext(script))
		{
			SYN_UNEXPECTED_EOF(script);
			goto err_exit;
		}
		if(name)
			memFree(name);
		name = strDup(lexLast(script));
		if(!evalOutput(name,script))
			goto err_exit;
	}
	result = true;
err_exit:;
	if(name)
		memFree(name);
	return result;
}

/* *********************************************************** */

V_LOCAL arg_p
evalSearchForFuncArg(vhandle_t script,int type)
{
	CALLER_NAME(evalSearchForFuncArg)
	arg_p arg;
	char *name;
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return NULL;
	}
	name = strDup(lexLast(script));
	arg = evalArgument(name,type,script);
	memFree(name);
	return arg;
}

V_LOCAL vboolean
ipl_width_of(vhandle_t script,vec_t *res)
{
	CALLER_NAME(ipl_width_of)
	arg_p arg = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!arg)
		return false;
	if(!arg->data.matrix)
	{
		if(arg->state.custom_width)
		{
			WARNING_GENERIC(ipl_using_state);
			*res = arg->state.custom_width;
			return true;
		}
		ERROR_GENERIC(ipl_null_func_arg);
		return false;
	}
	*res = arg->data.matrix->cols;
	return true;
}

V_LOCAL vboolean
ipl_height_of(vhandle_t script,vec_t *res)
{
	CALLER_NAME(ipl_height_of)
	arg_p arg = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!arg)
		return false;
	if(!arg->data.matrix)
	{
		if(arg->state.custom_height)
		{
			WARNING_GENERIC(ipl_using_state);
			*res = arg->state.custom_height;
			return true;
		}
		ERROR_GENERIC(ipl_null_func_arg);
		return false;
	}
	*res = arg->data.matrix->rows;
	return true;
}

V_LOCAL vboolean
ipl_size_of(vhandle_t script,vec_t *res)
{
	CALLER_NAME(ipl_size_of)
	arg_p arg = evalSearchForFuncArg(script,ARG_VECTOR);
	if(!arg)
		return false;
	if(!arg->data.vector)
	{
		if(arg->state.custom_width)
		{
			WARNING_GENERIC(ipl_using_state);
			*res = arg->state.custom_width;
			return true;
		}
		ERROR_GENERIC(ipl_null_func_arg);
		return false;
	}
	*res = arg->data.vector->size;
	return true;
}

V_LOCAL vboolean
ipl_depth_of(vhandle_t script,vec_t *res)
{
	CALLER_NAME(ipl_depth_of)
	arg_p arg = evalSearchForFuncArg(script,ARG_DECOMPOSITION);
	if(!arg)
		return false;
	if(!arg->data.decomposition)
	{
		ERROR_GENERIC(ipl_null_func_arg);
		return false;
	}
	*res = arg->data.decomposition->depth_levels;
	return true;
}

typedef vboolean (*ipl_func_t)(vhandle_t script,vec_t *res);

typedef struct	ipl_func_entry_s
{
	char_p 		*name_ptr;
	ipl_func_t	routine;
	char_p 		*desc_ptr;
}	ipl_func_entry_t, *ipl_func_entry_p;

V_LOCAL ipl_func_entry_t ipl_funcs[] =
{
	{
		&ipl_key_width_of,
		ipl_width_of,
		&ipl_desc_width_of,
	},
	{
		&ipl_key_height_of,
		ipl_height_of,
		&ipl_desc_height_of,
	},
	{
		&ipl_key_size_of,
		ipl_size_of,
		&ipl_desc_size_of,
	},
	{
		&ipl_key_depth_of,
		ipl_depth_of,
		&ipl_desc_depth_of,
	},
	{NULL,NULL,NULL}
};

V_LOCAL ipl_func_t
evalFindFunction(char *name)
{
	ipl_func_entry_p head;
	for(head = ipl_funcs;head->name_ptr;head++)
		if(STR_EQ(*head->name_ptr,name))
			return head->routine;
	return false;
}

void
evalHelpFuncs(void)
{
	ipl_func_entry_p head;
	conPuts("Available functions:\n");
	for(head = ipl_funcs;head->name_ptr;head++)
		conPrintf("\t\"%s\": %s%c\n",*head->name_ptr,*head->desc_ptr,
			(head + 1)->name_ptr ? ';' : '.');
}

/* *********************************************************** */

V_LOCAL vboolean
eval_last_real(vhandle_t script,vec_t *result)
{
	CALLER_NAME(eval_last_real)
	char *name;
	ipl_func_t func;
	if(!lexLastFloat(script,result)) /* check for constant */
	{
		name = lexLast(script);
		if(!name)
		{
			SYN_UNEXPECTED_EOF(script);
			return false;
		}
		func = evalFindFunction(name);
		if(!func)
		{
			SYN_UNDECLARED(script,name);
			return false;
		}
		if(!(*func)(script,result))
			return false;
	}
	return true;
}

vboolean
evalRealFunction(vhandle_t script,vec_t *result)
{
	lexNext(script);
	return eval_last_real(script,result);
}

vboolean
evalIntFunction(vhandle_t script,s32_t *result)
{
	CALLER_NAME(evalIntFunction)
	vec_t tmp, diff;
	if(lexNextInteger(script,result))
		return true;
	if(!eval_last_real(script,&tmp))
		return false;
	*result = (s32_t)floor(tmp);
	diff = (vec_t)*result - tmp;
	if(!IsZero(diff))
	{
		if(v_abs(diff) > 1)
		{
			PARSE_ERROR(script,ipl_roundoff_too_large);
			return false;
		}
		WARNING_GENERIC(ipl_cast_to_integer);
	}
	return true;
}

vboolean
evalUIntFunction(vhandle_t script,u32_t *result)
{
	CALLER_NAME(evalUIntFunction)
	s32_t itmp;
	vec_t tmp, diff;
	if(lexNextInteger(script,&itmp))
	{
		if(itmp < 0)
		{
			PARSE_ERROR(script,ipl_wrong_sign);
			return false;
		}
		*result = (u32_t)itmp;
		return true;
	}
	if(!eval_last_real(script,&tmp))
		return false;
	*result = (u32_t)floor(tmp);
	diff = (vec_t)*result - tmp;
	if(!IsZero(diff))
	{
		if(v_abs(diff) > 1)
		{
			PARSE_ERROR(script,ipl_roundoff_too_large);
			return false;
		}
		WARNING_GENERIC(ipl_cast_to_integer);
	}
	return true;
}

vboolean
evalBooleanFunction(vhandle_t script,vboolean *result)
{
	CALLER_NAME(evalBooleanFunction)
	char *token;
	vec_t tmp;
	token = lexNext(script);
	if(!token)
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	if(STR_EQ(token,ipl_key_true))
	{
		*result = true;
		return true;
	}
	if(STR_EQ(token,ipl_key_false))
	{
		*result = false;
		return true;
	}
	if(!eval_last_real(script,&tmp))
		return false;
	if(IsZero(tmp))
	{
		*result = false;
		return true;
	}
	if(IsZero(tmp - 1))
	{
		*result = true;
		return true;
	}
	WARNING_GENERIC(ipl_cast_to_integer);
	*result = true;
	return true;
}

/* *********************************************************** */

V_LOCAL matrix_p
check_state_create_matrix(arg_p arg)
{
	CALLER_NAME(check_state_create_matrix)
	V_ASSERT(arg->type == ARG_MATRIX);
	if(!arg->data.matrix)
	{
		eval_state_p st = &arg->state;
		if(!st->custom_width || !st->custom_height)
		{
			ERROR_GENERIC(ipl_state_wh_unset);
			return NULL;
		}
		arg->data.matrix = matAlloc(st->custom_height,st->custom_width);
		arg->owner = true;
	}
	return arg->data.matrix;
}

V_LOCAL vector_p
check_state_create_vector(arg_p arg)
{
	CALLER_NAME(check_state_create_vector)
	V_ASSERT(arg->type == ARG_VECTOR);
	if(!arg->data.vector)
	{
		eval_state_p st = &arg->state;
		if(!st->custom_width)
		{
			ERROR_GENERIC(ipl_state_size_unset);
			return NULL;
		}
		arg->data.vector = vecAlloc(st->custom_width);
		arg->owner = true;
	}
	return arg->data.vector;
}

/* *********************************************************** */

vp_stack_t fourier_transformers = TYPED_STACK_EMPTY_INITIALIZER;

V_LOCAL vhandle_t
evalFindFourierTransformer(u32_t rows,u32_t cols)
{
	CALLER_NAME(evalFindFourierTransformer)
	u32_t i, total, r, c;
	vhandle_t fft_trans, *cache;
	if(!IS_POWER_TWO(rows) || !IS_POWER_TWO(cols))
	{
		ERROR_GENERIC(ipl_fft_requires_powers2)
		return LIBV_INVALID_HANDLE;
	}
	cache = (vhandle_t*)vpsGetStorage(&fourier_transformers);
	total = vpsCount(&fourier_transformers);
	/* search in cache */
	for(i = 0;i < total;i++)
	{
		if(fftRVMGetSize(cache[i],&r,&c) &&
			(r == rows) && (c == cols))
			return cache[i];
	}
	/* nothing appropriate; create */
	fft_trans = fftRVMCreateContext(rows,cols);
	if(fft_trans != LIBV_INVALID_HANDLE)
		vpsPush(&fourier_transformers,fft_trans);
	return fft_trans;
}

void
evalDestroyAllFourierTransformers()
{
	vhandle_t fft_trans;
	while(vpsPop(&fourier_transformers,&fft_trans))
		if(fft_trans != LIBV_INVALID_HANDLE)
			fftRVMDestroyContext(fft_trans);
	vpsClear(&fourier_transformers);
}

/* *********************************************************** */

/*
 * Matrix operations
 */

V_LOCAL vboolean
matrix_copy_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_copy_op)
	char *token;
	arg_p ref_mat;
	if(arg->data.matrix)
	{
		WARNING_GENERIC(ipl_argument_override);
		if(arg->owner)
			memFree(arg->data.matrix);
		arg->data.matrix = NULL;
		arg->owner = false;
	}
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	token = strDup(lexLast(script));
	ref_mat = evalArgument(token,ARG_MATRIX,script);
	memFree(token);
	if(!ref_mat)
		return false;
	arg->data.matrix = matDuplicate(ref_mat->data.matrix);
	if(arg->data.matrix)
	{
		arg->owner = true;
		return true;
	}
	return false;
}

V_LOCAL vboolean
matrix_get_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_get_op)
	char *token;
	arg_p ref_mat;
	u32_t d_left, d_top;
	u32_t s_left, s_top, s_right, s_bottom;
	if(!evalUIntFunction(script,&d_left) ||
			!evalUIntFunction(script,&d_top))
		return false;
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	token = strDup(lexLast(script));
	ref_mat = evalArgument(token,ARG_MATRIX,script);
	memFree(token);
	if(!ref_mat)
		return false;
	if(!arg->data.matrix)
	{
		arg->data.matrix = matAlloc(ref_mat->data.matrix->rows,
			ref_mat->data.matrix->cols);
		arg->owner = true;
	}
	if(!evalUIntFunction(script,&s_left) ||
			!evalUIntFunction(script,&s_top) ||
			!evalUIntFunction(script,&s_right) ||
			!evalUIntFunction(script,&s_bottom))
		return false;
	matCopyElements(arg->data.matrix,ref_mat->data.matrix,
		d_left,d_top,s_left,s_top,s_right,s_bottom);
	return true;
}

V_LOCAL vboolean
matrix_resize_op(arg_p arg,vhandle_t script)
{
	/* CALLER_NAME(matrix_resize_op) */
	matrix_p m;
	vhandle_t resampler;
	u32_t new_width, new_height;
	if(!evalUIntFunction(script,&new_width) ||
			!evalUIntFunction(script,&new_height))
		return false;
	if(!arg->data.matrix)
		if(!check_state_create_matrix(arg))
			return false;
	m = arg->data.matrix;
	resampler = matCreateResampleContext(m->rows,m->cols,new_height,new_width);
	if(resampler == LIBV_INVALID_HANDLE)
		return false;
	m = matAlloc(new_height,new_width);
	matResample(resampler,m,arg->data.matrix);
	matDestroyResampleContext(resampler);
	if(arg->owner)
		memFree(arg->data.matrix);
	arg->data.matrix = m;
	arg->owner = true;
	return true;
}

V_LOCAL vboolean
matrix_input_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_input_op)
	char img_name[_MAX_PATH];
	V_ASSERT(arg->type == ARG_MATRIX);
	if(arg->data.matrix)
	{
		WARNING_GENERIC(ipl_argument_override);
		if(arg->owner)
			memFree(arg->data.matrix);
		arg->data.matrix = NULL;
		arg->owner = false;
	}
	if(!lexCopyNextToken(script,img_name,_MAX_PATH))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	arg->data.matrix = inputGetMatrix(&arg->state,img_name,lexNext(script));
	if(arg->data.matrix)
	{
		arg->owner = true;
		return true;
	}
	return false;
}

V_LOCAL vboolean
matrix_save_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_save_op)
	u32_t r, c, c_lim;
	vhandle_t outfile;
	matrix_p m;
	char file_name[_MAX_PATH];
	V_ASSERT(arg->type == ARG_MATRIX);
	if(!lexCopyNextToken(script,file_name,_MAX_PATH))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	outfile = fileOpenWrite(file_name);
	if(outfile == LIBV_INVALID_HANDLE)
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
	{
		fileClose(outfile);
		return false;
	}
	c_lim = m->cols - 1;
	for(r = 0;r < m->rows;r++)
	{
		for(c = 0;c < c_lim;c++)
			filePrintFloatStr(outfile,dotchar,m->cell[r][c],ipl_tab_string);
		filePrintFloatStr(outfile,dotchar,m->cell[r][c],ipl_eol_string);
	}
	fileClose(outfile);
	return true;
}

V_LOCAL vboolean
matrix_row_linear_op(arg_p arg,vhandle_t script)
{
	matrix_p m;
	vec_t min, max;
	if(!evalRealFunction(script,&min) || !evalRealFunction(script,&max))
		return false;
	if((m = check_state_create_matrix(arg)) == NULL)
		return false;
	matLinSpacedRows(m,min,max);
	return true;
}

V_LOCAL vboolean
matrix_col_linear_op(arg_p arg,vhandle_t script)
{
	matrix_p m;
	vec_t min, max;
	if(!evalRealFunction(script,&min) || !evalRealFunction(script,&max))
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	matLinSpacedCols(m,min,max);
	return true;
}

#include "argsused.h"
V_LOCAL vboolean
matrix_fourier_forward_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_fourier_forward_op)
	matrix_p m;
	vhandle_t context;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
		WARNING_GENERIC(ipl_already_frequency);
	context = evalFindFourierTransformer(m->rows,m->cols);
	if(context == LIBV_INVALID_HANDLE)
		return false;
	fftRVMForward(context,arg->data.matrix);
	arg->state.frequency_domain = true;
	return true;
}

#include "argsused.h"
V_LOCAL vboolean
matrix_fourier_inverse_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_fourier_inverse_op)
	matrix_p m;
	vhandle_t context;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(!arg->state.frequency_domain)
		WARNING_GENERIC(ipl_already_spatial);
	context = evalFindFourierTransformer(m->rows,m->cols);
	if(context == LIBV_INVALID_HANDLE)
		return false;
	fftRVMInverse(context,arg->data.matrix);
	arg->state.frequency_domain = false;
	return true;
}

V_LOCAL vboolean
matrix_zero_rows_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_zero_rows_op)
	matrix_p m;
	u32_t row_min, row_max;
	if(!evalUIntFunction(script,&row_min) ||
			!evalUIntFunction(script,&row_max))
		return false;
	if(row_min > row_max)
	{
		PARSE_ERROR(script,ipl_wrong_bounds)
		return false;
	}
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTZeroRows(m,row_min,row_max);
	}
	else
		matZeroRows(m,row_min,row_max);
	return true;
}

V_LOCAL vboolean
matrix_zero_cols_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_zero_cols_op)
	matrix_p m;
	u32_t col_min, col_max;
	if(!evalUIntFunction(script,&col_min) ||
			!evalUIntFunction(script,&col_max))
		return false;
	if(col_min > col_max)
	{
		PARSE_ERROR(script,ipl_wrong_bounds)
		return false;
	}
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTZeroCols(m,col_min,col_max);
	}
	else
		matZeroCols(m,col_min,col_max);
	return true;
}

V_LOCAL vboolean
matrix_band_pass_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_band_pass_op)
	matrix_p m;
	u32_t min, max;
	if(!evalUIntFunction(script,&min) ||
			!evalUIntFunction(script,&max))
		return false;
	if(min > max)
	{
		PARSE_ERROR(script,ipl_wrong_bounds)
		return false;
	}
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTBandPass(m,min,max);
	}
	else
		matBandPass(m,min,max);
	return true;
}

V_LOCAL vboolean
matrix_band_reject_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_band_reject_op)
	matrix_p m;
	u32_t min, max;
	if(!evalUIntFunction(script,&min) ||
			!evalUIntFunction(script,&max))
		return false;
	if(min > max)
	{
		PARSE_ERROR(script,ipl_wrong_bounds)
		return false;
	}
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTBandReject(m,min,max);
	}
	else
		matBandReject(m,min,max);
	return true;
}

typedef struct scale_cb_arg_s
{
	vec_t re_scale;
	vec_t im_scale;
}	scale_cb_arg_t, *scale_cb_arg_p;

#include "argsused.h"
V_LOCAL vboolean V_CALL
rvmft_scale_cb(void_p arg,u32_t row,u32_t col,vboolean negf,vec_t reim[])
{
	scale_cb_arg_p sa = (scale_cb_arg_p)arg;
	reim[0] *= sa->re_scale;
	reim[1] *= sa->im_scale;
	return true;
}

V_LOCAL vboolean
matrix_scale_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_scale_op)
	matrix_p m;
	vec_t scale;
	if(!evalRealFunction(script,&scale))
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		scale_cb_arg_t sa;
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		sa.re_scale = scale;
		sa.im_scale = scale;
		matRVMFTWalkCartesian(m,0,0,m->cols >> 1,m->rows >> 1,
			rvmft_scale_cb,&sa);
	}
	else
		matScale(m,m,scale);
	return true;
}

#include "argsused.h"
V_LOCAL vboolean V_CALL
rvmft_phase_shift_cb(void_p arg,u32_t row,u32_t col,vboolean negf,vec_t mph[])
{
	vec_t ofs = *((vec_p)arg);
	mph[1] += ofs;
	return true;
}

V_LOCAL vboolean
matrix_offset_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_offset_op)
	matrix_p m;
	vec_t offset;
	if(!evalRealFunction(script,&offset))
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTWalkPolar(m,0,0,m->cols >> 1,m->rows >> 1,
			rvmft_phase_shift_cb,&offset);
	}
	else
		matOffset(m,offset);
	return true;
}

V_LOCAL vboolean
matrix_mat_scale_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_mat_scale_op)
	matrix_p m;
	arg_p param = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!param)
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(!MAT_SAME_SIZE(m,param->data.matrix))
	{
		ERROR_GENERIC(ipl_size_mismatch);
		return false;
	}
	if(arg->state.frequency_domain != param->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_domain_mismatch);
		return false;
	}
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTMultiply(m,param->data.matrix);
	}
	else
		matScaleM(m,m,param->data.matrix);
	return true;
}

V_LOCAL vboolean
matrix_mat_unscale_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_mat_unscale_op)
	matrix_p m;
	arg_p param = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!param)
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(!MAT_SAME_SIZE(m,param->data.matrix))
	{
		ERROR_GENERIC(ipl_size_mismatch);
		return false;
	}
	if(arg->state.frequency_domain != param->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_domain_mismatch);
		return false;
	}
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTDivide(m,param->data.matrix);
	}
	else
		matUnscaleM(m,m,param->data.matrix);
	return true;
}

V_LOCAL vboolean V_CALL
rvmft_phase_add_cb(void_p arg,u32_t row,u32_t col,vboolean negf,
	vec_t mph1[],vec_t mph2[])
{
	mph1[1] += mph2[1];
	return true;
}

V_LOCAL vboolean
matrix_mat_offset_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_mat_offset_op)
	matrix_p m;
	arg_p param = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!param)
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	if(!MAT_SAME_SIZE(m,param->data.matrix))
	{
		ERROR_GENERIC(ipl_size_mismatch);
		return false;
	}
	if(arg->state.frequency_domain != param->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_domain_mismatch);
		return false;
	}
	if(arg->state.frequency_domain)
	{
		if(!IS_POWER_TWO(m->rows) || !IS_POWER_TWO(m->cols))
		{
			ERROR_GENERIC(ipl_fft_requires_powers2)
			return false;
		}
		matRVMFTWalkPolarPair(m,param->data.matrix,0,0,
			m->cols >> 1,m->rows >> 1,rvmft_phase_add_cb,NULL);
	}
	else
		matAdd(m,m,param->data.matrix);
	return true;
}

V_LOCAL vboolean
matrix_adjust_contrast_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_adjust_contrast_op)
	matrix_p m;
	vec_t min, max;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	if(!evalRealFunction(script,&min) ||
			!evalRealFunction(script,&max))
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	return matAdjustContrast(m,m,min,max);
}

V_LOCAL vboolean
matrix_normalize_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_normalize_op)
	matrix_p m;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	return matMaximizeContrast(m,m);
}

V_LOCAL vboolean
matrix_gamma_correction_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_gamma_correction_op)
	matrix_p m;
	arg_p param;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	param = evalSearchForFuncArg(script,ARG_VECTOR);
	if(!param)
		return false;
	m = check_state_create_matrix(arg);
	if(!m)
		return false;
	matGammaCorrection(m,m,param->data.vector,false);
	return true;
}

V_LOCAL vboolean
matrix_draw_decomp_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_draw_decomp_op)
	matrix_p m;
	arg_p param;
	bt_decomp_p dec;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	param = evalSearchForFuncArg(script,ARG_DECOMPOSITION);
	if(!param)
		return false;
	dec = param->data.decomposition;
	if(!arg->data.matrix)
	{
		arg->data.matrix = matAlloc(dec->source_rows,dec->source_cols);
		arg->owner = true;
	}
	m = arg->data.matrix;
	if(dec->source_rows != m->rows || dec->source_cols != m->cols)
	{
		PARSE_ERROR(script,ipl_size_mismatch);
		return false;
	}
    iplDrawDecomposition(m,dec);
	return true;
}

V_LOCAL vboolean
matrix_draw_decomp_scale_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_draw_decomp_op)
	matrix_p m;
	arg_p param, scale;
	bt_decomp_p dec;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	param = evalSearchForFuncArg(script,ARG_DECOMPOSITION);
	if(!param)
		return false;
	dec = param->data.decomposition;
	if(!arg->data.matrix)
	{
		arg->data.matrix = matAlloc(dec->source_rows,dec->source_cols);
		arg->owner = true;
	}
	m = arg->data.matrix;
	if(dec->source_rows != m->rows || dec->source_cols != m->cols)
	{
		PARSE_ERROR(script,ipl_size_mismatch);
		return false;
	}
	scale = evalSearchForFuncArg(script,ARG_VECTOR);
	if(dec->depth_levels != scale->data.vector->size)
	{
		PARSE_ERROR(script,ipl_size_mismatch);
		return false;
	}
    iplDrawDecompositionScale(m,dec,scale->data.vector);
	return true;
}

V_LOCAL vboolean
matrix_draw_decomp_linear_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(matrix_draw_decomp_linear_op)
	matrix_p m;
	arg_p param;
	bt_decomp_p dec;
	vec_t min, max;
	if(arg->state.frequency_domain)
	{
		ERROR_GENERIC(ipl_wrong_domain);
		return false;
	}
	param = evalSearchForFuncArg(script,ARG_DECOMPOSITION);
	if(!param)
		return false;
	dec = param->data.decomposition;
	if(!arg->data.matrix)
	{
		arg->data.matrix = matAlloc(dec->source_rows,dec->source_cols);
		arg->owner = true;
	}
	m = arg->data.matrix;
	if(dec->source_rows != m->rows || dec->source_cols != m->cols)
	{
		PARSE_ERROR(script,ipl_size_mismatch);
		return false;
	}
	if(!evalRealFunction(script,&min) ||
			!evalRealFunction(script,&max))
		return false;
    iplDrawDecompositionLinearScale(m,dec,min,max);
	return true;
}

V_LOCAL vec_p
eval_parse_list_of_reals(vhandle_t lexer,u32_t num_reals)
{
	CALLER_NAME(eval_parse_list_of_reals)
	u32_t i;
	vec_p dest;
	if(!lexNextIs(lexer,ipl_open_bracket))
	{
		SYN_EXPECTED(lexer,ipl_open_bracket);
		return NULL;
	}
	dest = (vec_p)memAlloc(num_reals * sizeof(vec_t));
	num_reals--;
	for(i = 0;i < num_reals;i++)
	{
		if(!evalRealFunction(lexer,dest + i))
			goto err_exit;
		if(!lexNextIs(lexer,ipl_comma))
		{
			SYN_EXPECTED(lexer,ipl_comma);
			goto err_exit;
		}
	}
	if(!evalRealFunction(lexer,dest + i))
		goto err_exit;
	if(!lexNextIs(lexer,ipl_close_bracket))
	{
		SYN_EXPECTED(lexer,ipl_close_bracket);
		goto err_exit;
	}
	return dest;
err_exit:;
	memFree(dest);
	return NULL;
}

V_LOCAL vboolean
matrix_transform_op(arg_p arg,vhandle_t script)
{
	vboolean res;
	u32_t idx, nargs;
	vec_p args;
	if(!iplFindTransform(IPL_MATRIX_TRANSFORM,lexNext(script),script,
			&idx,&nargs))
		return false;
	if(!check_state_create_matrix(arg))
		return false;
	if(nargs)
	{
		args = eval_parse_list_of_reals(script,nargs);
		if(!args)
			return false;
	}
	else
		args = NULL;
	res = iplExecuteTransform(idx,
		0,0,arg->data.matrix->cols - 1,arg->data.matrix->rows - 1,
		NULL,NULL,
		arg->data.matrix,NULL,
		args);
	if(args)
		memFree(args);
	return res;
}

V_LOCAL vboolean
cartesian_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

V_LOCAL vboolean
polar_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

V_LOCAL vboolean
matrix_pair_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

V_LOCAL vboolean
cartesian_pair_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

V_LOCAL vboolean
polar_pair_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

/* *********************************************************** */

/*
 * Vector operations
 */

V_LOCAL vboolean
vector_copy_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(vector_copy_op)
	char *token;
	arg_p ref_vec;
	if(arg->data.vector)
	{
		WARNING_GENERIC(ipl_argument_override);
		if(arg->owner)
			memFree(arg->data.vector);
		arg->data.vector = NULL;
		arg->owner = false;
	}
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	token = strDup(lexLast(script));
	ref_vec = evalArgument(token,ARG_VECTOR,script);
	memFree(token);
	if(!ref_vec)
		return false;
	arg->data.vector = vecDuplicate(ref_vec->data.vector);
	arg->owner = true;
	return true;
}

V_LOCAL vboolean
vector_save_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(vector_save_op)
	u32_t i, lim;
	vhandle_t outfile;
	vector_p v;
	char file_name[_MAX_PATH];
	if(!lexCopyNextToken(script,file_name,_MAX_PATH))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	outfile = fileOpenWrite(file_name);
	if(outfile == LIBV_INVALID_HANDLE)
		return false;
	v = check_state_create_vector(arg);
	if(!v)
		return false;
	lim = v->size - 1;
	for(i = 0;i < lim;i++)
		filePrintFloatStr(outfile,dotchar,v->elements[i],ipl_tab_string);
	filePrintFloatStr(outfile,dotchar,v->elements[i],ipl_eol_string);
	fileClose(outfile);
	return true;
}

V_LOCAL vboolean
vector_linear_op(arg_p arg,vhandle_t script)
{
	vector_p v;
	vec_t min, max;
	if(!evalRealFunction(script,&min) ||
			!evalRealFunction(script,&max))
		return false;
	v = check_state_create_vector(arg);
	if(!v)
		return false;
	vecLinSpaced(v,min,max);
	return true;
}

V_LOCAL vboolean
vector_histogram_op(arg_p arg,vhandle_t script)
{
	vector_p v;
	arg_p param = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!param)
		return false;
	v = check_state_create_vector(arg);
	if(!v)
		return false;
	matHistogram(v,param->data.matrix,false,NULL);
	return true;
}

V_LOCAL vboolean
vector_integrate_op(arg_p arg,vhandle_t script)
{
	vector_p v;
	v = check_state_create_vector(arg);
	if(!v)
		return false;
	vecRunningSum(v,v);
	return true;
}

V_LOCAL vboolean
vector_normalize_op(arg_p arg,vhandle_t script)
{
	vector_p v;
	v = check_state_create_vector(arg);
	if(!v)
		return false;
	vecMaximizeContrast(v,v);
	return true;
}

V_LOCAL vboolean
vector_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

V_LOCAL vboolean
vector_pair_transform_op(arg_p arg,vhandle_t script)
{
    return false;
}

/* *********************************************************** */

/*
 * Decomposition operations
 */

V_LOCAL vboolean
decomp_create_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(decomp_create_op)
	arg_p param;
	vec_t thresold;
	u32_t hst_divs;
	if(arg->data.decomposition)
	{
		WARNING_GENERIC(ipl_argument_override);
		if(arg->owner)
			iplFreeDecomposition(arg->data.decomposition);
		arg->data.decomposition = NULL;
		arg->owner = false;
	}
	param = evalSearchForFuncArg(script,ARG_MATRIX);
	if(!param)
		return false;
	if(!evalRealFunction(script,&thresold) ||
			!evalUIntFunction(script,&hst_divs))
		return false;
	if(hst_divs < 2)
	{
		PARSE_ERROR(script,ipl_incorrect_parameters);
		return false;
	}
	arg->data.decomposition =
		iplDecomposeImage(param->data.matrix,thresold,hst_divs);
	if(!arg->data.decomposition)
		return false;
	arg->owner = true;
	return true;
}

V_LOCAL vboolean
decomp_save_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(decomp_save_op)
	vboolean save_values, relative_frequencies;
	char fname[_MAX_PATH];
	if(!arg->data.decomposition)
	{
		PARSE_ERROR(script,ipl_incomplete_definition);
		return false;
	}
	if(!lexCopyNextToken(script,fname,_MAX_PATH))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	if(!evalBooleanFunction(script,&save_values) ||
			!evalBooleanFunction(script,&relative_frequencies))
		return false;
	iplSaveDecompositionDescription(fname,arg->data.decomposition,
		save_values,true,relative_frequencies);
	return true;
}

/* *********************************************************** */

/*
 * Operations common for matrices and vectors
 */

vboolean generic_constant_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(generic_constant_op)
	vec_t val, *ptr;
	u32_t r, c;
	vector_p v;
	matrix_p m;
	if(!evalRealFunction(script,&val))
		return false;
	switch(arg->type)
	{
		case ARG_MATRIX:
			m = check_state_create_matrix(arg);
			if(!m)
				return false;
			for(r = 0;r < m->rows;r++)
			{
				ptr = m->cell[r];
				for(c = 0;c < m->cols;c++)
					ptr[c] = val;
			}
			break;
		case ARG_VECTOR:
			v = check_state_create_vector(arg);
			if(!v)
				return false;
			ptr = v->elements;
			for(c = 0;c < v->size;c++)
				v->elements[c] = val;
			break;
		default:
			ERROR_GENERIC(ipl_invalid_operation);
			return false;
	}
	return true;
}

vboolean generic_uniform_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(generic_uniform_op)
	vec_t min, max;
	u32_t r;
	vector_p v;
	matrix_p m;
	if(!evalRealFunction(script,&min) ||
			!evalRealFunction(script,&max))
		return false;
	switch(arg->type)
	{
		case ARG_MATRIX:
			m = check_state_create_matrix(arg);
			if(!m)
				return false;
			v = vecAlloc(m->cols);
			for(r = 0;r < m->rows;r++)
			{
				ranNextVectorRange(random_gen,v,min,max);
				memCopy(m->cell[r],v->elements,v->size * sizeof(vec_t));
			}
			memFree(v);
			break;
		case ARG_VECTOR:
			v = check_state_create_vector(arg);
			if(!v)
				return false;
			ranNextVectorRange(random_gen,v,min,max);
			break;
		default:
			ERROR_GENERIC(ipl_invalid_operation);
			return false;
	}
	return true;
}

vboolean generic_gaussian_op(arg_p arg,vhandle_t script)
{
	CALLER_NAME(generic_uniform_op)
	vec_t mean, stdev;
	u32_t r;
	vector_p v;
	matrix_p m;
	if(!evalRealFunction(script,&mean) ||
			!evalRealFunction(script,&stdev))
		return false;
	switch(arg->type)
	{
		case ARG_MATRIX:
			m = check_state_create_matrix(arg);
			if(!m)
				return false;
			v = vecAlloc(m->cols);
			for(r = 0;r < m->rows;r++)
			{
				ranNextGaussianVector(random_gen,v,mean,stdev);
				memCopy(m->cell[r],v->elements,v->size * sizeof(vec_t));
			}
			memFree(v);
			break;
		case ARG_VECTOR:
			v = check_state_create_vector(arg);
			if(!v)
				return false;
			ranNextGaussianVector(random_gen,v,mean,stdev);
			break;
		default:
			ERROR_GENERIC(ipl_invalid_operation);
			return false;
	}
	return true;
}

/* *********************************************************** */

typedef vboolean (*op_routine_t)(arg_p,vhandle_t);

typedef struct op_parse_entry_s
{
	char 			*name;
	op_routine_t	routine;
	char			*desc;
}	op_parse_entry_t, *op_parse_entry_p;

V_LOCAL op_parse_entry_t matrix_operations[] =
{
	{
		"copy",
		matrix_copy_op,
		"copies contents of given matrix into current matrix"
	},
	{
		"input",
		matrix_input_op,
		"reads matrix from image file; requires image name and channel name"
	},
	{
		"get",
		matrix_get_op,
		"inserts at <left,top> rectangle from <ref_mat> with <left,top,right,bottom>"
	},
	{
		"resize",
		matrix_resize_op,
		"resizes the matrix to <new_width,new_height>"
	},
	{
		"save",
		matrix_save_op,
		"writes matrix to text file; requires file name"
	},
	{
		"constant",
		generic_constant_op,
		"sets all elements of matrix to given float constant"
	},
	{
		"uniform",
		generic_uniform_op,
		"uniform distribution; requires bounds"
	},
	{
		"gaussian",
		generic_gaussian_op,
		"normal distribution; requires mean and standard deviation"
	},
	{
		"row_linear",
		matrix_row_linear_op,
		"interpolates from given 'min' to given 'max', writes to every row"
	},
	{
		"col_linear",
		matrix_col_linear_op,
		"interpolates from given 'min' to given 'max', writes to every col"
	},
	{
		"fourier_forward",
		matrix_fourier_forward_op,
		"replaces matrix by its 2D forward Fourier transform"
	},
	{
		"fourier_inverse",
		matrix_fourier_inverse_op,
		"inverses Fourier transform"
	},
	{
		"scale",
		matrix_scale_op,
		"multiplies matrix elements by given real 'scale'"
	},
	{
		"offset",
		matrix_offset_op,
		"adds given 'offset' to matrix elements (or performs phase shift)"
	},
	{
		"mat_scale",
		matrix_mat_scale_op,
		"per-element multiplication of two matrices or two Fourier transforms"
	},
	{
		"mat_unscale",
		matrix_mat_unscale_op,
		"per-element division of two matrices or two Fourier transforms"
	},
	{
		"mat_offset",
		matrix_mat_offset_op,
		"per-element addition of two matrices or phase addition of two Fourier transforms"
	},
	{
		"zero_rows",
		matrix_zero_rows_op,
		"clears rows from given 'min' to given 'max' (can operate on Fourier transforms)"
	},
	{
		"zero_cols",
		matrix_zero_cols_op,
		"clears cols from given 'min' to given 'max' (can operate on Fourier transforms)"
	},
	{
		"band_reject",
		matrix_band_reject_op,
		"clears both cols and rows from given 'min' to given 'max' (can operate on Fourier transforms)"
	},
	{
		"band_pass",
		matrix_band_pass_op,
		"clears both cols and rows outside of given 'min' and 'max' (can operate on Fourier transforms)"
	},
	{
		"adjust_contrast",
		matrix_adjust_contrast_op,
		"linear map from given [min;max] to [0,1]"
	},
	{
		"normalize",
		matrix_normalize_op,
		"linear map from [matrix lower;matrix upper] to [0,1]"
	},
	{
		"gamma_correction",
		matrix_gamma_correction_op,
		"curve map from [matrix lower;matrix upper] to [0,1]; uses given vector as curve"
	},
    {
    	"draw_decomposition",
        matrix_draw_decomp_op,
        "draws given binary decomposition"
    },
    {
    	"draw_decomposition_scale",
        matrix_draw_decomp_scale_op,
        "draws given binary decomposition with given vector as color scale"
    },
    {
    	"draw_decomposition_linear",
        matrix_draw_decomp_linear_op,
        "draws given binary decomposition with linear scale from given min to given max"
    },
    {
    	"matrix_transform",
        matrix_transform_op,
        "executes spatial-domain transform"
    },
    {
    	"cartesian_transform",
        cartesian_transform_op,
        "executes frequency-domain transform"
    },
    {
    	"polar_transform",
        polar_transform_op,
        "executes frequency-domain polar transform"
    },
    {
    	"matrix_pair_transform",
        matrix_pair_transform_op,
        "executes spatial-domain transform for current and given matrices"
    },
    {
    	"cartesian_pair_transform",
        cartesian_pair_transform_op,
        "executes frequency-domain transform for current and given matrices"
    },
    {
    	"polar_pair_transform",
        polar_pair_transform_op,
        "executes frequency-domain polar transform for current and given matrices"
    },
	{NULL,NULL,NULL}
};

V_LOCAL op_parse_entry_t vector_operations[] =
{
	{
		"copy",
		vector_copy_op,
		"copies contents of given vector into current vector"
	},
	{
		"save",
		vector_save_op,
		"writes vector to text file; requires file name"
	},
	{
		"uniform",
		generic_uniform_op,
		"uniform distribution; requires bounds"
	},
	{
		"gaussian",
		generic_gaussian_op,
		"normal distribution; requires mean and standard deviation"
	},
	{
		"linear",
		vector_linear_op,
		"interpolates from given 'min' to given 'max'"
	},
	{
		"histogram",
		vector_histogram_op,
		"computes histogram of given matrix"
	},
	{
		"integrate",
		vector_integrate_op,
		"computes running sum (discrete integral)"
	},
	{
		"normalize",
		vector_normalize_op,
		"linear map from [vector lower;vector upper] to [0,1]"
	},
    {
    	"vector_transform",
        vector_transform_op,
        "executes spatial-domain transform"
    },
    {
    	"vector_pair_transform",
        vector_pair_transform_op,
        "executes spatial-domain transform for current and given vectors"
    },
	{NULL,NULL,NULL}
};

V_LOCAL op_parse_entry_t decomposition_operations[] =
{
	{
    	"decompose",
        decomp_create_op,
        "decomposes given matrix with given thresold and histogram_divs"
    },
	{
    	"save",
        decomp_save_op,
        "saves decomposition to given file; requires boolean 'save_values' and 'relative_frequencies' flags"
    },
	{NULL,NULL,NULL}
};

V_LOCAL op_routine_t
evalFindRoutine(char *name,op_parse_entry_p head)
{
	for(;head->name;head++)
		if(STR_EQ(head->name,name))
			return head->routine;
	return NULL;
}

/* *********************************************************** */

typedef struct toplevel_entry_s
{
	char	**key_ptr;
	u32_t	type;
	op_parse_entry_p operations;
	void	(*data_destructor)(void_p any_data);
	char	*desc;
}	toplevel_entry_t, *toplevel_entry_p;

V_LOCAL void
generic_free(void_p ptr)
{
	memFree(ptr);
}

V_LOCAL toplevel_entry_t toplevel_tokens[] =
{
	{
		&ipl_key_matrix,
		ARG_MATRIX,
		matrix_operations,
		generic_free,
		"definition of matrix"
	},
	{
		&ipl_key_vector,
		ARG_VECTOR,
		vector_operations,
		generic_free,
		"definition of vector"
	},
	{
		&ipl_key_decomposition,
		ARG_DECOMPOSITION,
		decomposition_operations,
		iplFreeDecomposition,
		"definition of binary image decomposition"
	},
	{NULL,0,0,NULL,NULL}
};

V_LOCAL void
evalListOperations(char *prefix,op_parse_entry_p head)
{
	for(;head->name;head++)
		conPrintf("%s\"%s\": %s%c\n",prefix,head->name,head->desc,
		(head + 1)->name ? ';' : '.');
}

void evalHelp()
{
	toplevel_entry_p head = toplevel_tokens;
	conPuts("Known top-level block types:\n");
	for(;head->key_ptr;head++)
	{
		conPrintf("\"%s\": %s;\nValid operations:\n",*head->key_ptr,
			head->desc);
		evalListOperations(ipl_tab_string,head->operations);
	}
	evalHelpFuncs();
}

V_LOCAL toplevel_entry_p
evalToplevelEntry(u32_t type)
{
	toplevel_entry_p head = toplevel_tokens;
	for(;head->key_ptr;head++)
		if(head->type == type)
			return head;
	return NULL;
}

arg_p
evalArgument(char *name,int type,vhandle_t script)
{
	CALLER_NAME(evalArgument)
	arg_p arg = NULL;
	op_routine_t op_routine;
	toplevel_entry_p top;
	char *token = NULL;
	arg = evalSearchForName(name);
	if(arg)
	{
		if(arg->type != type)
		{
			ERROR_GENERIC_2ARG(ipl_name_in_use,name);
			return NULL;
		}
		if(!arg->data.any)
		{
			ERROR_GENERIC_2ARG(ipl_circular_reference,name);
			return NULL;
		}
		return arg;
	}
	top = evalToplevelEntry(type);
	V_ASSERT(top);
	arg = (arg_p)memAllocZ(sizeof(arg_t));
	V_ASSERT(arg);
	stateDefault(&(arg->state));
    arg->id = strDup(name);
    arg->type = type;
    arg->destroy = top->data_destructor;
    arg->owner = false;
	lexPush(script,LEX_POSITION_BIT);
	if(!scriptSearchPair(script,*top->key_ptr,arg->id))
		goto err_exit;
	if(!lexNextIs(script,ipl_open_brace))
	{
		SYN_EXPECTED(script,ipl_open_brace);
		goto err_exit;
	}
	while(lexNext(script))
	{
		if(token)
			memFree(token);
		token = strDup(lexLast(script));
		if(STR_EQ(token,ipl_close_brace))
			goto normal_exit;
		if(STR_EQ(token,ipl_key_state))
		{
			if(stateChange(&(arg->state),script))
				continue;
			break;
		}
		op_routine = evalFindRoutine(token,top->operations);
		if(op_routine)
		{
			if((*op_routine)(arg,script))
				continue;
			break;
		}
		PARSE_ERROR(script,ipl_unknown_operation);
		break;
	}
err_exit:;
	if(arg)
	{
		evalFreeArg(arg);
		arg = NULL;
	}
normal_exit:;
	if(arg && !arg->data.any)
	{
		PARSE_ERROR(script,ipl_incomplete_definition);
		evalFreeArg(arg);
		arg = NULL;
	}
	if(token)
		memFree(token);
	lexPop(script,LEX_POSITION_BIT);
	if(arg)
		vpsPush(&cached_args,arg);
    return arg;
}
